import asyncio
import datetime
import logging
from dataclasses import dataclass, field
from inspect import iscoroutinefunction as isAsync
from typing import Callable, Dict, List

import httpx
from apscheduler.schedulers.asyncio import AsyncIOScheduler

logger = logging.getLogger("Poster")
logger.setLevel(logging.INFO)
if not logger.handlers:
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            "[Poster][%(asctime)s][%(levelname)s][%(filename)s:%(lineno)d]: %(message)s", "%H:%M:%S"
        )
    )
    logger.addHandler(handler)


def countLog(name: str = "", start: int = 0):
    "计数装饰器"
    def inner(fn):
        if start < 0:
            return fn
        times = start

        async def wapper(*arg, **kwargs):
            nonlocal times
            logger.info("%s 第 %d 次轮询", name, times)
            times += 1
            return await fn(*arg, **kwargs)
        return wapper
    return inner


@dataclass
class User:
    level: int
    xp: int
    uid: int
    url: str
    watch: List[str]


@dataclass
class Post:
    mid: str
    time: int
    text: str
    type: str
    source: str

    uid: str
    name: str
    face: str
    pendant: str
    description: str

    follower: str
    following: str

    attachment: List[str]
    picUrls: List[str]

    repost: "Post"

    @classmethod
    def transform(cls: "Post", post: dict):
        "将输入 parse 的数据字典转为 Post 格式"
        return post

    @classmethod
    def parse(cls: "Post", post: dict) -> "Post":
        "递归解析"
        if post is None or len(post) == 0:
            return None
        post = cls.transform(post)
        return Post(repost=cls.parse(post.pop("repost")), **post)

    @property
    def date(self) -> str:
        "返回规定格式字符串时间"
        return datetime.datetime.fromtimestamp(self.time).strftime("%H:%M:%S")

    @property
    def data(self) -> dict:
        "返回可以以 data 发送至后端的数据格式"
        res = dict(self.__dict__)
        if self.repost is not None:
            res["repost"] = self.repost.data
        return res


@dataclass
class Comments:
    post: str
    comments: Dict[str, Post] = field(default_factory=dict)


class Request:
    "通用请求类"

    BaseHeaders = {
        "Connection": "keep-alive",
        "Accept-Language": "zh-CN,zh;q=0.9",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
    }

    def __init__(self, headers: dict = BaseHeaders, cookies: str = ""):
        if cookies:
            headers.update({"cookie": cookies})
        self.session = httpx.AsyncClient(headers=headers)

    async def request(self, method, url, *args, **kwargs):
        try:
            res = await self.session.request(method, url, *args, **kwargs)
            assert res.status_code == 200, "请求失败"
            return res.json()
        except Exception as e:
            logger.error(e)
            return None


class Poster(Request):
    """
    usage:

    async with Poster(uid, token, url) as poster:
        await poster.update(post)

    or

    poster = Poster(uid, token, url)

    await poster.login()

    await poster.update(post)
    """

    scheduler = AsyncIOScheduler(timezone="Asia/Shanghai")
    run_time = lambda seconds: datetime.datetime.now() + datetime.timedelta(seconds=seconds)

    def __init__(self, uid: int, token: str, baseurl: str):
        self.uid = uid
        self.token = token
        self.baseurl = baseurl
        self.__vaild = False

        super().__init__()

    async def __aenter__(self): return await self.login()

    async def __aexit__(self, type, value, trace): ...

    async def login(self) -> "Poster":
        "登录"
        try:
            res = await self.session.get(f"{self.baseurl}/login", params={
                "uid": self.uid,
                "token": self.token,
            })
            data = res.json()
            if data["code"] == 0:
                self.__vaild = True
                self.users = [User(**u) for u in data["data"]]
                for user in self.users:
                    if user.uid == self.uid:
                        self.me = user
                        logger.info(f"用户 {user.uid} LV{user.level}({user.xp}/100) 登录成功")
            else:
                raise Exception(data["data"])
        except Exception as e:
            logger.error(e)
            self.__vaild = False
        return self

    async def update(self, post: Post):
        "增"
        if self.__vaild:
            res = await self.session.post(f"{self.baseurl}/update", params={"token": self.token}, json=post.data)
            data = res.json()
            if data["code"] != 2:
                logger.info(data["data"])
        else:
            logger.error("未登录")

    async def post(self, beginTs: int = 0, endTs: int = -1):
        "查"
        res = await self.session.get(f"{self.baseurl}/post", params={"beginTs": int(beginTs), "endTs": int(endTs)})
        data = res.json()
        if data["code"] == 0:
            return data["data"]
        else:
            logger.error(data["data"])
            return []

    async def modify(self, user: User) -> User:
        "改"
        try:
            res = await self.session.post(f"{self.baseurl}/modify", params={
                "uid": self.uid,
                "token": self.token,
            }, data=user.__dict__)
            data = res.json()
            if data["code"] == 0:
                return User(**data["data"])
            else:
                raise Exception(data["data"])
        except Exception as e:
            logger.error(e)
            return None

    @classmethod
    def add_job(cls: "Poster", fn, start: int = 0, interval: int = 5, name: str = "", count: int = -1, args: list = list(), kwargs: dict = dict()):
        @cls.scheduler.scheduled_job("interval", next_run_time=cls.run_time(start), seconds=interval, args=args, kwargs=kwargs)
        @countLog(name, count)
        async def wapper(*arg, **kwargs):
            return await fn(*arg, **kwargs)

    @classmethod
    def job(cls: "Poster", start: int = 0, interval: int = 5, name: str = "", count: int = -1, args: list = list(), kwargs: dict = dict()):
        "轮询装饰器"
        def inner(fn):
            @cls.scheduler.scheduled_job("interval", next_run_time=cls.run_time(start), seconds=interval, args=args, kwargs=kwargs)
            @countLog(name, count)
            async def wapper(*arg, **kwargs):
                return await fn(*arg, **kwargs)
            return wapper
        return inner

    @staticmethod
    async def stepBrother():
        "I'm STUCK"
        while True:
            await asyncio.sleep(60)

    @classmethod
    def run(cls: "Poster", fn: Callable, *posters: "Poster"):
        """
        fn: 阻塞函数 如果没有可以传入 Poster.stepBrother

        *posters: 提交器 会调用对应的 login() 方法
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        for poster in posters:
            loop.create_task(poster.login())
        cls.scheduler.start()
        if isAsync(fn):
            loop.run_until_complete(fn())
        else:
            fn()



from json import loads
from typing import Callable, Coroutine
from dataclasses import dataclass

import httpx
from aiowebsocket.converses import AioWebSocket
from bilibili_api.utils.AsyncEvent import AsyncEvent

from .utils import logger


@dataclass
class RoomInfo:
    room_id: int
    uid: int
    title: str
    name: str
    cover: str
    user_face: str
    user_description: str


class Receive:
    "异步接收"

    def __init__(self, recv: Callable):
        self.recv = recv

    def __aiter__(self):
        return self

    async def __anext__(self):
        return loads(await self.recv())


class BiliGo(AsyncEvent):
    """
    连接 biligo-ws-live 的适配器
    """

    def __init__(self, aid: str, url: str, *listening_rooms):
        super().__init__()
        #  接入 biligo-ws-live 时的 id 用来区分不同监控程序
        self.aid = aid
        # biligo-ws-live 运行地址
        self.url = url
        # 将监听房间号告知 biligo-ws-live
        httpx.post(self.url+'/subscribe', headers={"Authorization": self.aid}, data={'subscribes': list(listening_rooms)})

    def on(self, event_name: str) -> Callable:
        """
        装饰器注册事件监听器。

        Args:
            event_name (str): 事件名。
        """
        def decorator(func: Coroutine):
            async def wapper(args):
                return await func(*args)

            self.add_event_listener(event_name, wapper)
            return func

        return decorator

    async def run(self):
        """
        阻塞异步连接
        """

        async with AioWebSocket(self.url + f"/ws?id={self.aid}") as aws:
            logger.info('Adapter 连接成功')
            async for evt in Receive(aws.manipulator.receive):
                roomInfo = RoomInfo(**evt["live_info"])
                self.dispatch(evt['command'], (roomInfo, evt["content"]))



from datetime import datetime, timedelta
from typing import List, Dict

from lxml import etree

from .utils import Post, Request, Comments


class WeiboPost(Post):
    @staticmethod
    def created_at(timeText: str) -> int:
        "解析微博时间字段转为时间戳"
        return int(datetime.strptime(timeText, "%a %b %d %H:%M:%S %z %Y").timestamp())

    @classmethod
    def transform(cls: "WeiboPost", mblog: dict):
        user: dict = mblog.get("user", {})
        if user is None: user = dict()
        return {
            "mid": str(mblog["mid"]),
            "time": cls.created_at(mblog["created_at"]),
            "text": mblog["text"],
            "type": "weibo",
            "source": mblog["source"],

            "uid": str(user.get("id")),
            "name": user.get("screen_name"),
            "face": user.get("avatar_hd"),
            "pendant": "",
            "description": user.get("description"),

            "follower": str(user.get("followers_count")),
            "following": str(user.get("follow_count")),

            "attachment": [],
            "picUrls": [p["large"]["url"] for p in mblog.get("pics", [])],
            "repost": mblog.get("retweeted_status")
        }

class WeiboComment(Post):
    @staticmethod
    def created_at(timeText: str) -> int:
        """
        标准化微博发布时间

        参考 https://github.com/Cloud-wish/Dynamic_Monitor/blob/main/main.py#L575
        """
        created_at = datetime.now()
        
        if u"分钟" in timeText:
            minute = timeText[:timeText.find(u"分钟")]
            minute = timedelta(minutes=int(minute))
            created_at -= minute
        elif u"小时" in timeText:
            hour = timeText[:timeText.find(u"小时")]
            hour = timedelta(hours=int(hour))
            created_at -= hour
        elif u"昨天" in timeText:
            created_at -= timedelta(days=1)
        elif timeText.count('-') == 1:
            created_at -= timedelta(days=365)    
        
        return int(created_at.timestamp())

    @classmethod
    def transform(cls: "WeiboComment", com: dict):
        user: dict = com.get("user", {})
        if user is None: user = dict()
        pic: str = com.get("pic", {}).get("large", {}).get("url", None)
        return {
            "mid": str(com["id"]),
            "time": WeiboComment.created_at(com["created_at"]),
            "text": com["text"],
            "type": "weiboComment",
            "source": com["source"],

            "uid": str(user.get("id")),
            "name": user.get("screen_name"),
            "face": user.get("profile_image_url"),
            "pendant": "",
            "description": "",

            "follower": str(user.get("followers_count")),
            "following": str(user.get("friends_count")),

            "attachment": com["attachment"],
            "picUrls": [pic] if pic else [],
            "repost": com["repost"]
        }


class WeiboRequest(Request):
    def __init__(self, cookies: str):
        super().__init__(cookies=cookies)
        self.usersComments: Dict[str, Comments] = dict()

    async def posts(self, uid: int | str):
        data = await self.request("GET", f"https://m.weibo.cn/api/container/getIndex?containerid=107603{uid}")
        if data is None:
            return
        for card in data["data"]["cards"][::-1]:
            if card["card_type"] != 9: continue
            yield WeiboPost.parse(card["mblog"])

    async def comments(self, post: WeiboPost):
        data = await self.request("GET", f"https://m.weibo.cn/api/comments/show?id={post.mid}")
        if data is None:
            return

        # 清空储存的评论
        postName = post.type + post.mid
        if self.usersComments.get(post.uid, None) is None:
            self.usersComments[post.uid] = Comments(postName)
        elif self.usersComments[post.uid].post != postName:
            self.usersComments[post.uid] = Comments(postName)

        for com in data["data"]["data"][::-1]:
            mid = str(com["id"])
            if mid in self.usersComments[post.uid].comments: continue

            com["attachment"] = [postName]
            com["repost"] = self.usersComments[post.uid].comments.get(str(com.get("reply_id", "")), None)
            self.usersComments[post.uid].comments[mid] = com
            yield WeiboComment.parse(com)


def parse_text(text: str):
    "获取纯净博文内容"
    span = etree.HTML(f'<div id="post">{text}</div>')
    # 将表情替换为图片链接
    for _span in span.xpath('//div[@id="post"]/span[@class="url-icon"]'):
        alt = _span.xpath('./img/@alt')[0]
        src = _span.xpath('./img/@src')[0]
        _span.insert(0, etree.HTML(f'<p>{alt}: {src}</p>'))

    # 获取这个 span 的字符串形式 并去除 html 格式字符
    text: List[str] = [p.replace(u'\xa0', '').replace('&#13;', '\n') for p in span.xpath('.//text()')]

    # 记录所有 <a> 标签出现的位置
    apos: List[int] = [0]
    for _a in span.xpath('.//a/text()'):
        try:
            apos.append(text.index(_a, apos[-1]))
        except ValueError:
            ...
    else:
        apos.pop(0)
    return text, apos

def get_content(texts: List[str]) -> str:
    return "".join([t.split(": ")[0] if ": " in t else t for t in texts])


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r", encoding="utf8") as f:
    requires = f.read()

setup(
    name="weibo-poster",
    version="0.0.2",
    license="MIT",
    author="Drelf2018",
    author_email="drelf2018@outlook.com",
    description="对应 weibo-webhook 的提交器",
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=requires.splitlines(),
    keywords=['python', 'weibo', 'webhook'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    
    url="https://github.com/Drelf2018/weibo-poster",
    python_requires=">=3.8",
)